/*
    This file is part of vekfont.

    vekfont is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    vekfont is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    Contributed by Georg-Johann Lay <georgjohann@web.de>
*/

#ifndef VEKTOR_ZEICHEN_H
#define VEKTOR_ZEICHEN_H

#include <inttypes.h>

// ber dieses Makro kann die Breite der verwendeten Variablen
// eingestellt werden. Fr minimalen Codeverbrauch etwa auf AVR
// und bei einer Anzeige, deren x- und y-Koordinaten 8 Bit breit
// sind, gengen 8 Bits. Fr grere Anzeigen oder Host-Darstellung
// kann es erforderlich sein, mit mehr als 8 Bit zu rechnen.
#ifndef VF_INT_BITS
#define VF_INT_BITS 8
#endif

//////////////////////////////////////////////////////////////////////

// VEKFONT enthlt mehrere, im folgenden beschriebene Konfigurations-
// Bits, mit denen die Eigenschaften des Zeichensatzes eingestellt
// werden knnen. Den geringsten Speicherplatzbedarf hat
// VEKFONT=0:  Nur Grobuchstaben, keine Abbildung von Umlauten/Klein-
//             buchstaben, etc.
// VEKFONT=-1: Bietet die beste Untersttzung bei grerem Code
#ifndef VEKFONT
#define VEKFONT (0)
//#define VEKFONT (-1)
#endif

// Bit von VEKFONT
// Enthlt der Zeichensatz deutsche Umlaute , , , , ,  und
// die Ligatur ?
#define VF_UMLAUTS       (1 << 0)

// Bit von VEKFONT
// Enthlt der Zeichensatz Kleinbuchstaben?
#define VF_LOWERCASE     (1 << 1)

// Bit von VEKFONT
// Enthlt der Zeichensatz Pfeil-Symbole (ansprechbar via <, >, _, ^)
#define VF_ARROWS        (1 << 2)

// Bit von VEKFONT
// Enthlt der Zeichensatz Punktuation wie (+, -, *, ?, ...)
#define VF_PUNKT         (1 << 3)

// Bit von VEKFONT
// Falls der Zeichensatz keine Umlaute enthlt: Sollen Umlaute wie
//  als AE dargestellt werden?
#define VF_MAP_UMLAUTS   (1 << 4)

// Bit von VEKFONT
// Falls der Zeichensatz keine Kleinbuchstaben enthlt:
// Sollen Kleinbuchstaben als Grobuchstaben dargestellt werden?
#define VF_MAP_LOWERCASE (1 << 5)

// Bit von VEKFONT
// Kleinbuchstaben 'e' und 's' ragen 1 Pixel ber die Mittellinie
// und werden daher zu gro. Wird dieses Flag gesetzt, dann
// werden diese Buchstaben bei .zoom > 1 etwas geschrumpft,
// damit sie die Hhe der anderen Kleinbuchstaben erhalten.
#define VF_SHRINKY       (1 << 5)

// Bit von VEKFONT
// Wenn dieses Bit gesetzt ist, dann kann zur Laufzeit zwischen
// Fixed-Font und Proportionalschrift umgestellt werden.
// Auf den Platzbedarf des Zeichensatzes hat das keinen Einfluss,
// das Programm wird aber etwas langsamer und grer.
#define VF_PROPORTIONAL (1 << 6)

//////////////////////////////////////////////////////////////////////
 
// Falls LOOKUP_END > LOOKUP_START ist, dann wird zum schnellen
// Auffinden der Zeicheninformation eine Lookup-Tabelle im RAM
// aufgebaut. Die Tabelle enthlt dann Indizes der Zeichen mit
// ASCII-Code von VF_LOOKUP_START bis VF_LOOKUP_END.
// Der RAM-Bedarf in Bytes ist
//      2 * ( VF_LOOKUP_END - VF_LOOKUP_START + 1 )
#ifndef VF_LOOKUP_START
#define VF_LOOKUP_START '*'
#endif // LOOKUP_START

#ifndef VF_LOOKUP_END
#define VF_LOOKUP_END 0
#endif // LOOKUP_END

// Fr Sonderanwendung, wird normal nicht gebraucht
// Beim Aufruf wird u.U. kein Pixel ausgegeben
#ifndef VF_PLOTTER
#define VF_PLOTTER 0
#endif

#if VF_INT_BITS == 8
typedef uint8_t vf_uint_t;
typedef int8_t  vf_int_t;
#elif VF_INT_BITS == 16
typedef uint16_t vf_uint_t;
typedef int16_t  vf_int_t;
#else
typedef unsigned int vf_uint_t;
typedef signed   int vf_int_t;
#endif

// Wird ausgegeben, wenn ein Zeichen nicht gefunden wird.
// Kennzeichnet ausserdem das Tabellenende.
#define VF_UNKNOWN ((uint8_t) '\\')

// Feld mit der Zeichensatz-Information
extern const uint8_t vektor_zeichen_data[];

// (un)signed Pixel
typedef struct { vf_uint_t x, y; } vf_upixel_t;
typedef struct { vf_int_t x, y;  } vf_spixel_t;

// Struktur mit der Zeichen-Information
typedef struct
{
    // RAM-Adresse des auszugebenden Textes
    const char * text;

    // Wird zur Ausgabe eines Pixels aufgerufen
    void (*beam_xy)(vf_uint_t x, vf_uint_t y);
    
    // Wird Aufgerufen zwischen zwei nicht-nebeneinander liegenden
    // Pixeln, also wenn der Strahl/Stift einen Sprung machen muss.
    void (*beam_skip)(void);

    // Startpunkt des auszugebenden Textes:
    // die untere linke Ecke des ersten Zeichens
    vf_upixel_t start;

    // Zoom-Faktor
    uint8_t zoom;

    // Proportional-Schrift oder Fixed-Font?
    // Nur wirksam zusammen mit VF_PROPORTIONAL
    uint8_t proportional;

    ///////////////////////////////////////////////////////////////////
    // Interne Verwendung
    ///////////////////////////////////////////////////////////////////
    const uint8_t * addr;   // Adresse ins vektor_zeichen_data[] Feld
    vf_spixel_t  d;         // Abstand zum nchsten Pixel
    vf_uint_t nPixels;      // n Pixel in dieser Linie
    vf_uint_t iPixel;       // Index des Pixels in dieser Linie
    vf_uint_t itext;        // Index des momentanen Zeichens in text[]
    vf_upixel_t p;          // Zeichen-Position
    vf_upixel_t pix;        // Pixel-Position
    uint8_t   nLines;       // Linien in diesem Zeichen
    uint8_t   iLine;        // Index der momentanen Linie
    uint8_t   noskip;       // ersten Pixel zeichnen
    uint8_t   right;        // nach rechts gehen bei Fixed
    uint8_t   left;         // nach links  gehen bei Proportional
    uint8_t   inject_char;  // Extra-Zeichen bei Konvertierung ->ae etc.
    uint8_t   zeichen;      // Das aktuelle Zeichen
    uint8_t   shrinky;      // y-Ausdehnung auf 3/4 schrumpfen
    vf_upixel_t shrink;     // Pixel-Position
} letter_drawinfo_t;

extern uint8_t draw_letters (letter_drawinfo_t * di);

// len dxy start end
// 765 432 11111 000
#define LINE(len,dx,dy,start,end)  \
    ((len << 5)                    \
    | (start ? 2 : 0)              \
    | (end   ? 1 : 0)              \
    | ((                           \
    (dx ==  1 && dy ==  0) ? 4 :   \
    (dx ==  1 && dy == -1) ? 3 :   \
    (dx ==  0 && dy == -1) ? 2 :   \
    (dx == -1 && dy == -1) ? 1 :   \
    (dx == -1 && dy ==  0) ? 0 :   \
    (dx == -1 && dy ==  1) ? 7 :   \
    (dx ==  0 && dy ==  1) ? 6 : 5 \
    ) << 2))

#define DOTS_L LINE(7,-1,-1,0,0)
#define DOTS_R LINE(7,-1,-1,1,0)
#define DOTS_B LINE(7,-1,-1,0,1)
#define DOTS_T LINE(7,-1,-1,1,1)

#ifdef __i386__

typedef struct
{
    vf_spixel_t min, max;
} boundbox_t;

// Bresenham-Algorithmus um gerade Linien zu zeichnen.
#define BL_DRAW_START (1 << 0)
#define BL_DRAW_END   (1 << 1)

typedef struct
{
    void (*beam_xy) (vf_uint_t x, vf_uint_t y);
    // Momentane Koordinaten
    vf_upixel_t p;

    // Start- und End-Koordinaten
    vf_upixel_t start, end;

    // dd. ist Diagonalschritt
    int8_t ddx, ddy;

    // pd. ist Parallelschritt
    int8_t pdx, pdy;

    // Fehlerschritte schnell, langsam
    // el ist auch Anzahl zu zeichnender Pixel
    vf_int_t es, el;

    // Fehler
    vf_int_t err;

    // Nummer des Pixels
    vf_uint_t pix;
    
    // Anzahl Pixels
    vf_uint_t npix;

    // ersten Pixel zeichnen?
    uint8_t draw_start;
} bresenham_line_t;


extern uint16_t count_pixels (char c, uint8_t zoom);
extern void boundbox (char c, uint8_t zoom, boundbox_t *bb);
extern void draw_reset (letter_drawinfo_t * di);
extern void bresh_init (bresenham_line_t * bl,
                        void (*beam_xy) (vf_uint_t, vf_uint_t),
                        vf_upixel_t start, vf_upixel_t end,
                        uint8_t start_end);

extern uint8_t bresh (bresenham_line_t * bl);

#endif // __i386__
#endif // VEKTOR_ZEICHEN_H
